/**
 * @file default.h
 * @author Joe Wingbermuehle
 * @date 2017
 *
 * @brief Default configuration.
 *
 */

#ifndef DEFAULT_H
#define DEFAULT_H

/** Base configuration that is always included first. */
extern const char * const BASE_CONFIG;

/** A default configuration that is included if no configuration is found. */
extern const char * const DEFAULT_CONFIG;

#endif /* DEFAULT_H */
